package Stanford::LDAP::RegressionTesting;

## no critic (Subroutines::RequireArgUnpacking);
## no critic (Documentation::RequirePodAtEnd);
## no critic (NamingConventions);
## no critic (InputOutput::RequireCheckedSyscalls);

use strict;
use warnings;
use autodie;

=head1 NAME

Stanford::LDAP::RegressionTesting - Support module for Stanford OpenLDAP regression testing

=cut

# Make this a Moose class
use Moose;
use namespace::autoclean;

# Run tests against an LDAP server.

use Data::Dumper;
use Carp;
use Authen::SASL;
use Net::LDAP qw(LDAP_SUCCESS LDAP_SIZELIMIT_EXCEEDED);
use Test::More;
use Time::Local;

use Stanford::Orange::Util qw/ run_command_improved trim slurp /;
use Stanford::FlexTest;
use Stanford::FlexTest::Collection;

# Hostname of LDAP server.
has server => (
    is       => 'rw',
    isa      => 'Str',
    required => 1,
);

# Make sure the server name is fully-qualified.
around 'server' => sub {
    my $orig = shift;
    my $self = shift;

    if ($self->{'server'}) {
        $self->{'server'} = normalize_server($self->{'server'});
    }
};

has ldap => (
    is       => 'rw',
    isa      => 'Any',
    required => 0,
);

has debug => (
    is       => 'rw',
    isa      => 'Bool',
    required => 1,
    default  => 0,
);

has ticket_cache => (
    is       => 'rw',
    isa      => 'Str',
    required => 0,
    default  => q{},
);

has using_anonymous_bind => (
    is       => 'rw',
    isa      => 'Bool',
    required => 1,
    default  => 0,
);

has using_simple_bind => (
    is       => 'rw',
    isa      => 'Bool',
    required => 1,
    default  => 0,
);

has simple_bind_dn => (
    is       => 'rw',
    isa      => 'Str',
    required => 0,
);

has simple_bind_password => (
    is       => 'rw',
    isa      => 'Str',
    required => 0,
);

has simple_bind_password_wallet_name => (
    is       => 'rw',
    isa      => 'Str',
    required => 0,
);

has simple_bind_password_file_name => (
    is       => 'rw',
    isa      => 'Str',
    required => 0,
);

# If set to true will invoke the start_tls method when binding.
has start_tls => (
    is       => 'rw',
    isa      => 'Bool',
    required => 1,
    default  => 0,
);

has in_nagios_mode => (
    is       => 'rw',
    isa      => 'Bool',
    required => 1,
    default  => 0,
);

my $CACHE        = 1;
my $CACHE_ENABLE = 0;

=head2 Methods

=over

=item $self->ldap_bind

Bind the object to the LDAP server in the "server" property. Takes into
account whether the authentication is GSSAPI or simple-bind.

Returns the Net::LDAP created.
=cut

sub ldap_bind {
    my $self = shift;

    $self->progress('entered bind_to_ldap');

    my $server = $self->server();

    my $ldapuri = $self->ldapuri();
    my $ldap    = Net::LDAP->new($ldapuri)
      or croak "error creating Net::LDAP object at '$ldapuri': $@";

    # Are we binding anonymously. If so, do it know and leave.
    if ($self->using_anonymous_bind()) {
        $self->progress('binding anonymously');
        $ldap->bind();
        $self->progress("finished binding to server $server");
        $self->ldap($ldap);
        return $ldap;
    }

    # We don't want to STARTTLS if we already bind directly using the
    # (discouraged) ldaps URI.
    if ($self->using_ssl() && $self->start_tls()) {
        croak 'cannot STARTTLS is already using ldaps://';
    }

    # Do we invoke STARTTLS? If so, do it now. Note that we add the option
    # "verify => 'require'" so that the start_tls method will die unless the
    # the entire valid certificate chain is sent by the LDAP server.
    if ($self->start_tls()) {
        my $result = $ldap->start_tls(verify => 'require',);
        #if ($result && ($result->{'resultCode'})) {
        #    die "STARTTLS failed: " . $result->errorMessage();
        #}
        $self->progress('using TLS for this connection');
    } else {
        $self->progress('NOT using TLS for this connection');
    }

    my ($dn, $password);

    $self->progress("about to bind to server $server");

    my $status;
    if ($self->using_simple_bind()) {
        $self->progress('setting up a simple bind connection');
        $dn       = $self->simple_bind_dn();
        $password = $self->simple_bind_password();

        $status = $ldap->bind($dn, password => $password);
    } elsif ($self->ticket_cache()) {
        $self->progress('binding with ticket cache ' . $self->ticket_cache());
        ## no critic (Variables::RequireLocalizedPunctuationVars);
        $ENV{'KRB5CCNAME'} = $self->ticket_cache();

        my $sasl = Authen::SASL->new(mechanism => 'GSSAPI');

        # $ldap->sasl_parms('sasl' => $sasl);

        $status = $ldap->bind('sasl' => $sasl);
    } else {
        my $sasl = Authen::SASL->new(mechanism => 'GSSAPI');
        $status = $ldap->bind('sasl' => $sasl);
    }

    if (!$status) {
        if (!$ldap) {
            $ldap->unbind;
        }
        croak "ERROR Bind error connecting to $server ";
    }

    $self->progress("finished binding to server $server");

    $self->ldap($ldap);
    return $ldap;
}

=item $self->to_string

Generate a string representation of $self.

=cut

sub to_string {
    my $self = shift;

    my $server    = $self->server();
    my $ldapuri   = $self->ldapuri();
    my $debug     = $self->debug();
    my $simple    = $self->using_simple_bind();
    my $start_tls = $self->start_tls();
    my $using_ssl = $self->using_ssl();

    my $simple_formatted    = $simple    ? 'YES' : 'NO';
    my $start_tls_formatted = $start_tls ? 'YES' : 'NO';
    my $using_ssl_formatted = $using_ssl ? 'YES' : 'NO';

    return <<"EOQ";
ldapuri:      $ldapuri
server:       $server
debug:        $debug
simple bind?: $simple_formatted
start_tls?:   $start_tls_formatted
using_ssl?:   $using_ssl_formatted
EOQ
}

=item $self->ldapuri

Return an ldap URI appropriate to the settings.

=cut

sub ldapuri {
    my $self = shift;

    my $server = $self->server();

    my $ldapuri;
    if ($self->using_simple_bind()) {
        $ldapuri = "ldaps://$server";
    } else {
        $ldapuri = "ldap://$server";
    }

    return $ldapuri;
}

=item $self->using_ssl

Return 1 if the ldap URI starts with "ldaps", 0 otherwise.

=cut

sub using_ssl {
    my $self = shift;

    my $ldapuri = $self->ldapuri();
    if ($ldapuri =~ m{ldaps}xsm) {
        return 1;
    } else {
        return 0;
    }
}

=item $self->progress

Print to stdout extra messages. Only prints out if the debug property
evaluates to true.

=cut

sub progress {
    my $self = shift;

    my ($msg) = @_;
    if ($self->debug()) {
        print "$msg\n";
    }
    return;
}

=item $self->problem

Print to stdout messages about problems. Only prints out if the debug property
evaluates to true.

=cut

sub problem {
    my $self = shift;

    my ($msg) = @_;
    if ($self->debug()) {
        print "problem: $msg\n";
    }
    return;
}

=item $self->ldap_search

Do an LDAP search. The usual parameters of base, scope, filter, and attrs
are accepted. The parameter sizelimit is also accepted to limit the size of
the result set returned.

Returns the results as a hash reference.

=cut

sub ldap_search {
    my $self = shift;
    my %args = (
        'base'      => 'dc=stanford,dc=edu',
        'scope'     => 'sub',
        'sizelimit' => '10',
        'attrs'     => [],
        @_,
    );

    my $base      = $args{'base'};
    my $scope     = $args{'scope'};
    my $filter    = $args{'filter'};
    my $attrs     = $args{'attrs'};
    my $sizelimit = $args{'sizelimit'};

    my $attrs_string;
    if ($attrs) {
        $attrs_string = join(q{,}, @{$attrs});
    } else {
        $attrs_string = 'ALL';
    }

    $self->progress("about to search on base: $base, "
          . "filter: $filter, scope: $scope, attrs: $attrs_string, "
          . "sizelimit: $sizelimit");

    my $result = $self->ldap()->search(
        'base'      => $base,
        'scope'     => $scope,
        'filter'    => $filter,
        'sizelimit' => $sizelimit,
        'attrs'     => $attrs,
    );

    my $msg;
    if ($result->is_error()) {
        my $code = $result->code();
        if ($code == LDAP_SIZELIMIT_EXCEEDED) {
            $msg = 'with LDAP_SIZELIMIT_EXCEEDED (this can be ignored)';
            $self->progress("search ended: $msg");
        } else {
            $msg = "in error; code=$code";
            $self->problem("search ended: $msg");
        }
    } else {
        $msg = 'without error';
        $self->progress("search ended: $msg");
    }

    return $result->as_struct();
}

=item $self->get_entry_attributes

Given a DN return as a hash reference all that entry's attributes.

=cut

# Returns a hash mapping attribute name to an array reference of values.
# Will return the empty hash if no results.
sub get_entry_attributes {
    my $self = shift;
    my ($dn) = @_;

    # Split dn.
    my ($prefix, $suffix);
    if ($dn =~ m{^([^,]+),(.*)$}xsm) {
        $prefix = $1;
        $suffix = $2;
    }
    my $results = $self->ldap_search(
        'filter' => $prefix,
        'base'   => $suffix,
        'scope'  => 'sub',
    );

    my $attributes = $results->{$dn};

    my %key_to_value = ();
    if ($attributes) {
        %key_to_value = %{$attributes};
    } else {
        %key_to_value = ();
    }

    return %key_to_value;
}

=item $self->add_fake_person

Create a person object in the cn=test,dc=stanford,dc=edu tree.
Set's the cn and sn attributes.

=cut

sub add_fake_person {
    my $self = shift;
    my ($basedn) = @_;

    # Generate a random cn
    my $cn = gen_random(10);
    my $dn = "cn=$cn,$basedn";
    $self->progress("dn is $dn");

    # Generate a random sn
    my $sn = gen_random(12);

    my $attributes_ref = [
        'objectClass' => ['person'],
        'cn'          => $cn,
        'sn'          => $sn,
    ];

    my $msgid = $self->ldap->add($dn, 'attrs' => [@{$attributes_ref}]);

    return (@{$attributes_ref}, 'dn' => $dn);
}

=item $self->delete_entry

Delete the entry with the supplied $dn. We only allow deletion in the test
tree.

=cut

sub delete_entry {
    my $self = shift;

    my ($dn) = @_;

    if (!$dn) {
        croak 'no dn supplied';
    } elsif ($dn !~ m{cn=test,dc=stanford,dc=edu$}xsm) {
        croak "can only delete entries in the test tree: $dn";
    } else {
        $self->ldap->delete($dn);
    }

    return;
}

=item $self->delete_attribute_value

Given an entry's DN, an attribute, and a value, delete all that entry's attributes
matching that attribute and value.


=cut

sub delete_attribute_value {
    my $self = shift;

    my ($dn, $attribute, $value) = @_;

    my $result;
    if (!$dn) {
        croak 'no dn supplied';
    } else {
        $result
          = $self->ldap->modify($dn, delete => { $attribute => [$value] });
    }

    return $result;
}

=item gen_random

Generate a random string of letters and numbers of specified length.

=cut

sub gen_random {
    my ($length) = @_;

    my @chars = ('a' .. 'z', '0' .. '9');
    my $string;

    for my $i (1 .. $length) {
        $string .= $chars[rand @chars];
    }

    return $string;
}

=item $self->read_simple_bind_password

Read the simple bind password stored in the wallet pointed to by the
simple_bind_password_wallet_name property. Store the password in
the simple_bind_password property.

Returns nothing.

=cut

# Read the simple bind password from the wallet object.
sub read_simple_bind_password {
    my $self = shift;

    my $password_file_name = $self->simple_bind_password_file_name();

    my $password;
    if ($password_file_name) {
        $self->progress("reading password from file '$password_file_name'");

        # Get the first line of the password file.
        my $lines = slurp($password_file_name);
        my @lines = split(/\n/xsm, $lines);
        $password = $lines[0];
    } else {
        my $wallet_name = $self->simple_bind_password_wallet_name();
        $self->progress("reading password from wallet object '$wallet_name'");
        $password = $self->read_wallet($wallet_name);
    }

    $self->simple_bind_password($password);

    return;
}

=item read_wallet

Read the supplied wallet object into a string. Return that string.
=cut

sub read_wallet {
    my $self = shift;

    my ($wallet_name) = @_;

    if (!$wallet_name) {
        croak 'cannot read from wallet without a wallet object name';
    }

    #<<<  do not let perltidy touch this
    my @cmd = (
        'wallet',
        'get',
        'file',
        $wallet_name,
        );
    #>>>

    my ($stdout, $stderr, $rv) = run_command_improved(@cmd);
    if ($stderr) {
        croak "error extracting form wallet: $!";
    }

    $stdout = trim($stdout);

    return $stdout;
}

=item $self->use_simple_bind

Set the flag that says we are using simple bind authentication.

=cut

sub use_simple_bind {
    my $self = shift;

    $self->read_simple_bind_password();
    $self->using_simple_bind(1);

    return 1;
}

=item $self->get_all_uids

Get all the uids from the accounts tree associated with entries with an
active suEmailStatus attribute. If $sizelimit is sepcified only return
$sizelimit many uids.

=cut

sub get_all_uids {
    my $self = shift;

    my ($sizelimit) = @_;

    my $uids_aref;
    if ($CACHE_ENABLE) {
        $uids_aref = $CACHE->get('uids');
    }

    my @all_uids = ();
    if (!$CACHE_ENABLE || (not defined $uids_aref)) {
        $self->progress('cache MISS');
        my $results = $self->ldap_search(
            'filter'    => 'suEmailStatus=active',
            'attrs'     => ['uid'],
            'base'      => 'cn=accounts,dc=stanford,dc=edu',
            'sizelimit' => $sizelimit,
        );
        foreach my $entry (values %{$results}) {
            my $uid = $entry->{'uid'}->[0];
            push(@all_uids, $uid);
        }

        $uids_aref = \@all_uids;

        if ($CACHE_ENABLE) {
            $self->progress('setting cache with key uids');
            $CACHE->set('uids', $uids_aref, '30 minutes');
        }
    } else {
        @all_uids = @{$uids_aref};
    }

    return @all_uids;
}

=item get_attributes_uid

Return all the attributes for a person with a given uid.

=cut

sub get_attributes_uid {
    my ($self) = shift;

    my ($uid, $attrs_aref) = @_;

    if (!defined($attrs_aref)) {
        $attrs_aref = [];
    }

    my $results_href = $self->ldap_search(
        'filter' => "uid=$uid",
        'attrs'  => $attrs_aref,
    );

    return $results_href;
}

=item $self->change_attribute

Given a DN an attribute and one more values, change that attribute to the
supplied values.

Returns nothing.

=cut

sub change_attribute {
    my ($self) = shift;

    my ($DN, $attribute, $values) = @_;

    my $mesg = $self->ldap->modify(
        $DN,
        replace => {
            $attribute => $values
        }
    );
    return;
}

=item $self->add_attribute

Given a DN, an attribute, and a value, add that attribute/value to the entry.

Returns nothing.

=cut

sub add_attribute {
    my ($self) = shift;

    my ($DN, $attribute, $value) = @_;

    my $mesg = $self->ldap->modify(
        $DN,
        add => {
            $attribute => $value
        }
    );
    return $mesg;
}

=item $self->delete_attributes

Given an entry DN and an attribute name, delete all that entry's
attributes matching that name.

Returns nothing.

=cut

sub delete_attributes {
    my ($self) = shift;

    my ($DN, $attribute) = @_;

    my $mesg = $self->ldap->modify($DN, delete => [$attribute]);
    return;
}

=item in_suPrivilegeGroup

Return 1 if the user with uid in the supplied suPrivilegegroup in the
supplied tree.

=cut

sub in_suPrivilegeGroup {
    my ($self) = shift;

    my ($uid, $suPrivilegeGroup, $tree) = @_;

    my $results_href = $self->ldap_search(
        'base'   => "cn=$tree,dc=stanford,dc=edu",
        'filter' => "(&(uid=$uid)(suPrivilegeGroup=$suPrivilegeGroup))",
        'attrs'  => ['uid'],
    );

    my @dns = keys %{$results_href};

    if (scalar(@dns) == 1) {
        my $value = $results_href->{ $dns[0] };
        if (exists($value->{'uid'})) {
            return 1;
        } else {
            return 0;
        }
    } elsif (scalar(@dns) > 1) {
        croak "found more than user with the uid '$uid'!";
    } else {
        return 0;
    }
}

=item get_dn_from_uid

Given a the DN for a given uid. Use the supplied $tree ('people' or
'accounts').

=cut

sub get_dn_from_uid {
    my ($self) = shift;

    my ($uid, $tree) = @_;

    my $results_href = $self->ldap_search(
        'base'   => "cn=$tree,dc=stanford,dc=edu",
        'filter' => "uid=$uid",
        'attrs'  => ['uid'],
    );

    my @dns = keys %{$results_href};

    if (scalar(@dns) == 1) {
        return $dns[0];
    } elsif (scalar(@dns) > 1) {
        croak "found more than user with the uid '$uid'!";
    } else {
        croak "no '$tree' tree entry with $uid found";
    }
}

=item get_dn_from_cn

Given a CN and a base, return the DN. If more than one result is found
raise an exception. If not result is found, reais an exception.

=cut

sub get_dn_from_cn {
    my ($self) = shift;

    my ($cn, $base) = @_;

    my $results_href = $self->ldap_search(
        'base'   => $base,
        'filter' => "cn=$cn",
        'attrs'  => ['cn'],
    );

    my @dns = keys %{$results_href};

    if (scalar(@dns) == 1) {
        return $dns[0];
    } elsif (scalar(@dns) > 1) {
        croak "found more than entry with the cn '$cn'!";
    } else {
        croak "no '$base' entry with cn '$cn' found";
    }
}

=item $self->add_user_to_privgroup

# Add the user (dn) to the supplied suPrivilegeGroup.

Returns nothing.

=cut

sub add_user_to_privgroup {
    my ($self) = shift;

    my ($dn, $suprivilegegroup) = @_;

    my $mesg = $self->ldap->modify(
        $dn,
        add => {
            'suPrivilegeGroup' => $suprivilegegroup,
        }
    );

    return $mesg;
}

=item remove_user_from_privgroup

Remove the user with $dn from the supplied suPrivilegeGroup.

=cut

sub remove_user_from_privgroup {
    my ($self) = shift;

    my ($dn, $suprivilegegroup) = @_;

    my $mesg = $self->ldap->modify(
        $dn,
        delete => {
            'suPrivilegeGroup' => $suprivilegegroup,
        }
    );

    return $mesg;
}

=item $self->add_user_to_group

Adds a user to a cn=groups,dc=stanford,dc=edu group entry.

Returns nothing.

=cut

sub add_user_to_group {
    my ($self) = shift;

    my ($user_dn, $group_dn) = @_;
    return;
}

=item compare_arrays

Return 1 if the two arrays are the same, 0 otherwise.

=cut

sub compare_arrays {
    my ($aref1, $aref2) = @_;

    if (!defined($aref1) && !defined($aref2)) {
        return 1;
    } elsif (!defined($aref1) || !defined($aref2)) {
        return 0;
    } else {
        my $a1 = join(q{,}, sort @{$aref1});
        my $a2 = join(q{,}, sort @{$aref2});

        return ($a1 eq $a2);
    }
}

=item normalize_server

Add '.stanford.edu' to the server name if not fully qualified. This
is necessary so that if we do a STARTTLS the certificate name (which
will have '.stanford.edu' at the end) matches the server name we
supply.

=cut

sub normalize_server {
    my ($server) = @_;

    # 1. If the name contains even a single ".", do nothing.
    if ($server =~ m{[.]}xsm) {
        return $server;
    }

    # 2. This name is not fully-qualified. So we add '.stanford.edu'.
    # But we have to be careful of the port.
    my ($hostname, $port) = split(/:/xsm, $server);

    if ($port) {
        return "${hostname}.stanford.edu:$port";
    } else {
        return "${hostname}.stanford.edu";
    }
}

=item try_several_times

Attempt the supplied function reference several times.

=cut

sub try_several_times {
    my ($self) = shift;

    my ($fref, $try_limit, $sleep) = @_;

    if (!$try_limit) {
        $try_limit = 3;
    }

    if (!defined($sleep)) {
        $sleep = 2;
    }

    my $number_tries = 0;
    while ($number_tries < $try_limit) {
        my $result = $fref->();
        if ($result) {
            return 1;
        } else {
            ++$number_tries;
            sleep($sleep);
        }
    }

    return 0;
}

=item $self->test_read_base

Run the "base" test returning a flextests array.

=cut

sub test_read_base {
    my ($self) = shift;

    my ($result, $fmsg);
    my $flextests = Stanford::FlexTest::Collection->new();
    $flextests->ok_mode(!$self->in_nagios_mode);

    my $results = $self->ldap_search(
        'base'   => q{},
        'scope'  => 'base',
        'filter' => 'objectclass=*',
        'attrs'  => [],
    );

    my $r1 = $results->{q{}};

    my @keys = keys %{$r1};

    # Test 1. Did we find something?
    $result = scalar(@keys) > 0;
    $fmsg   = 'found something in the base';
    $flextests->add($result, $fmsg);

    # Test 2. If we found something, it should be an objectClass.
    if (scalar(@keys) > 0) {
        $result = (lc($keys[0]) eq 'objectclass');
        $fmsg   = 'found objectClass in base';
        $flextests->add($result, $fmsg);
    }

    return $flextests;
}

=item $self->test_quick_read

Run the "quick" test returning a flextests array.

=cut

sub test_quick_read {
    my ($self) = shift;

    my ($result, $fmsg);
    my $flextests = Stanford::FlexTest::Collection->new();
    $flextests->ok_mode(!$self->in_nagios_mode);

    my $filter    = 'suName=*';
    my $attrs_ref = ['suName'];
    my $base      = 'cn=accounts,dc=stanford,dc=edu';
    my $sizelimit = 100;

    my $search_result = $self->ldap_search(
        'base'      => $base,
        'filter'    => $filter,
        'attrs'     => $attrs_ref,
        'sizelimit' => $sizelimit
    );

    my %dn_to_attrs = %{$search_result};

    my @dns          = keys %dn_to_attrs;
    my $number_found = scalar(@dns);
    $self->progress("the search found $number_found suName attributes");

    # Test 1. Did we find enought elements?
    $result = (scalar(@dns) == $sizelimit);
    $fmsg   = "found at least $sizelimit elements";
    $flextests->add($result, $fmsg);

    return $flextests;
}

=item $self->test_uids

Run the "uids" test returning a flextests array.

=cut

sub test_uids {
    my ($self) = shift;

    my ($result, $fmsg);
    my $flextests = Stanford::FlexTest::Collection->new();
    $flextests->ok_mode(!$self->in_nagios_mode);

    my $expected_number_of_uids;
    if ($self->using_simple_bind()) {
        $expected_number_of_uids = 450;
    } else {
        $expected_number_of_uids = 3000;
    }

    my @UIDS = $self->get_all_uids($expected_number_of_uids);

    $self->progress('found ' . scalar(@UIDS) . ' accounts');

    # Test 1. Did we find enought uids?
    $result = scalar(@UIDS) >= $expected_number_of_uids;
    $fmsg   = "at least $expected_number_of_uids accounts found";
    $flextests->add($result, $fmsg);

    # Pick some uids at random and get some more accounts information.
    for my $i (1 .. 10) {
        my $uid = @UIDS[rand @UIDS];

        # Get account information on this user.
        my $dn = "uid=$uid,cn=accounts,dc=stanford,dc=edu";

        my %key_to_values = $self->get_entry_attributes($dn);

        my $owner_aref = $key_to_values{'owner'};
        if ($owner_aref) {
            $self->progress("accounts entry for $uid has an owner");
        } else {
            $self->progress("accounts entry for $uid does NOT have an owner");
        }

        # Test 2N. Did we find enought uids?
        $result = $owner_aref;
        $fmsg   = "accounts entry for $uid has an owner";
        $flextests->add($result, $fmsg);

        my $owner = $owner_aref->[0];
        $self->progress("owner is '$owner'");

        # Is the owner in the people tree?
        if ($owner =~ m{cn=people}xsm) {

            # YES.
            $self->progress('suName is ' . $key_to_values{'suname'});
            $result = length($key_to_values{'suname'}->[0]);
            $fmsg   = "uid $uid has a suName of sufficient length";
            $flextests->add($result, $fmsg);

            # Do some extra tests if we are NOT using simple bind
            if (!$self->using_simple_bind()) {
                ## It turns out that in some of the environments there
                ## are users with no suPrivilegeGroup's. So, we cannot
                ## use that as a test.
                #my $priv_groups = $key_to_values{'suprivilegegroup'};
                #
                #$result = $priv_groups;
                #$fmsg   = "uid $uid is in at least one suPrivilegeGroup";
                #$flextests->add($result, $fmsg);
                #if ($result) {
                #    $self->progress(
                #        "uid $uid is in at least one suPrivilegeGroup");
                #} else {
                #    $self->progress(
                #        "uid $uid isn't in at least one suPrivilegeGroup");
                #}
                #
                #my @priv_groups = @{ $key_to_values{'suprivilegegroup'} };
                #$self->progress("uid $uid has these suPrivilegeGroups: "
                #      . join(q{,}, @priv_groups));
                #
                #$result = (scalar(@priv_groups) > 0);
                #$fmsg   = "uid $uid is in at least one suPrivilegeGroup";
                #$flextests->add($result, $fmsg);

                my %people_k2v = $self->get_entry_attributes($owner);

                if (exists($people_k2v{'ou'})) {
                    my $ou = $people_k2v{'ou'}->[0];
                    $result = (length($ou) > 1);
                    $fmsg   = "uid $uid has an ou";
                } else {
                    $result = 1;
                    $fmsg   = "uid $uid has no ou";
                }
                $flextests->add($result, $fmsg);

                # Believe it or not, not everyone has an sn, at
                # least not in all the enironments.
                if (exists($people_k2v{'sn'})) {
                    my $sn = $people_k2v{'sn'}->[0];
                    $result = length($sn) > 1;
                    $fmsg   = "uid $uid has an sn";
                } else {
                    $result = 1;
                    $fmsg   = "uid $uid has no sn";
                }
                $flextests->add($result, $fmsg);
            }
        }
    }

    return $flextests;
}

=item $self->test_replication

Run the "uids" test returning a flextests array.
This set of tests writes to the test tree (cn=test) and verifies that
the results are replicated. The $self is the MASTER while the
$ldap_master must be a Stanford::LDAP::RegressionTesting object pointing
to the master.

=cut

sub test_replication {
    my ($ldap_master) = shift;    # $self is the MASTER

    my ($ldap_replica_binds_aref, $delay) = @_;

    $ldap_master->progress('master is ' . $ldap_master->server());

    my @ldap_replicas = @{$ldap_replica_binds_aref};

    my $flextests = Stanford::FlexTest::Collection->new();
    $flextests->ok_mode(!$ldap_master->in_nagios_mode);

    my $base = 'cn=test,dc=stanford,dc=edu';

    # ### #    # ### #    # ### #    # ### #    # ### #    # ### #
    my $add_delete_fake_person_fref = sub {

        my ($result, $fmsg);
        # STEP 1: Add fake person to MASTER
        my %attributes = $ldap_master->add_fake_person($base);
        my $dn         = $attributes{'dn'};

        $result = $dn;
        $fmsg   = "new fake person has dn $dn";
        $flextests->add($result, $fmsg);

        $ldap_master->progress(
            "added fake person '$dn' to master " . $ldap_master->server());

        Time::HiRes::usleep(400 * 1000);
        $ldap_master->progress("sleeping an extra $delay seconds");
        sleep $delay;    # extra sleep

        # STEP 2: Check that fake person appears in all the REPLICAS.
        foreach my $ldap_replica (@ldap_replicas) {
            my $replica_name = $ldap_replica->server();
            my %key_to_value = $ldap_replica->get_entry_attributes($dn);

            $ldap_replica->progress(
                "verifying $dn replicated to $replica_name");

            my $number_of_keys = scalar(keys(%key_to_value));
            $ldap_replica->progress(
                "number of $dn attributes found: $number_of_keys");

            if ($number_of_keys <= 0) {
                $ldap_replica->problem('no attributes found');
            }

            my $main_result;
            $main_result = ($number_of_keys > 0);
            $fmsg        = "got some entry attributes for $dn";
            $flextests->add($main_result, $fmsg);

            $result = ($main_result
                  && ($key_to_value{'sn'}->[0] eq $attributes{'sn'}));
            $fmsg = "sn matches in replica $replica_name";
            $flextests->add($result, $fmsg);
            if ($main_result && !$result) {
                $ldap_replica->problem("$fmsg failed");
            }

            $result = ($main_result
                  && ($key_to_value{'cn'}->[0] eq $attributes{'cn'}));
            $fmsg = "cn matches in replica $replica_name";
            $flextests->add($result, $fmsg);
            if ($main_result && !$result) {
                $ldap_replica->problem("$fmsg failed");
            }
        }

        # Step 3. Delete fake name from MASTER.
        $ldap_master->delete_entry($dn);
        $ldap_master->progress(
            "deleted fake person '$dn' to master " . $ldap_master->server());
        Time::HiRes::usleep(400 * 1000);
        sleep $delay;    # extra sleep

        # Step 4. Verify that the fake name got deleted from REPLICAS.
        foreach my $ldap_replica (@ldap_replicas) {
            my $replica_name = $ldap_replica->server();
            my %key_to_value
              = $ldap_replica->get_entry_attributes($attributes{'dn'});

            $ldap_replica->progress(
                "verifying deleted from replica $replica_name");

            $result = (scalar(keys(%key_to_value)) == 0);
            $fmsg   = "entry $dn deleted on replica $replica_name";
            $flextests->add($result, $fmsg);
            if (!$result) {
                $ldap_replica->problem("$fmsg failed");
            }
        }
    };
    # ### #    # ### #    # ### #    # ### #    # ### #    # ### #

    for my $i (1 .. 3) {
        $add_delete_fake_person_fref->();
    }

    return $flextests;
}

=item $self->replica_seconds_behind_master

Returns the number of seconds the replica is behind the master.  We do
this by comparing the contextCSN's of the "data" tree (dc=stanford,dc=edu)
of the master and the replica.

=cut

sub replica_seconds_behind_master {
    my ($ldap_master) = shift;    # $self is the MASTER

    my ($ldap_replica) = @_;

    my $replica_name = $ldap_replica->server();

    # Get the contextCSN's for the master and this replica.
    #<<<  do not let perltidy touch this
    my %master_sid_to_datetime
      = $ldap_master->get_contextCSNs('dc=stanford,dc=edu');
    my %replica_sid_to_datetime
      = $ldap_replica->get_contextCSNs('dc=stanford,dc=edu');
    #>>>

    # Get the SIDs from the replica's contextCSNs.
    my @replica_sids = keys %replica_sid_to_datetime;

    my ($max_diff, $max_diff_sid);
    foreach my $sid (@replica_sids) {
        if (exists($master_sid_to_datetime{$sid})) {
            my $replica_datetime = $replica_sid_to_datetime{$sid}->[0];
            my $master_datetime  = $master_sid_to_datetime{$sid}->[0];
            my $replica_epoch    = $replica_sid_to_datetime{$sid}->[1];
            my $master_epoch     = $master_sid_to_datetime{$sid}->[1];

            my $diff = $master_epoch - $replica_epoch;
            $ldap_master->progress(
                "replica is $diff seconds behind master ($sid)");

            if (!defined($max_diff) || ($diff > $max_diff)) {
                $max_diff     = $diff;
                $max_diff_sid = $sid;
            } else {
                # Nothing: $max_diff stays the same
            }
        } else {
            $ldap_master->progress(
                "no master contextCSN matching replica SID '$sid'");
        }
    }

    return ($max_diff_sid, $max_diff);
}

=item $self->test_replication_delay

Test that the replicas are not too far behind the master in their syncing.

=cut

sub test_replication_delay {
    my ($ldap_master) = shift;    # $self is the MASTER

    my ($ldap_replica_binds_aref, $sync_delay) = @_;

    $ldap_master->progress('master is ' . $ldap_master->server());

    my @ldap_replicas = @{$ldap_replica_binds_aref};

    my $flextests = Stanford::FlexTest::Collection->new();
    $flextests->ok_mode(!$ldap_master->in_nagios_mode);

    foreach my $ldap_replica (@ldap_replicas) {
        my $replica_name = $ldap_replica->server();
        my ($sid, $diff)
          = $ldap_master->replica_seconds_behind_master($ldap_replica);

        # Give it one more chance in case the master had not had a change for a
        # long time but suddenly had a change right before this test ran.
        if ($diff > $sync_delay) {
            my $sleep_time = 20;
            $ldap_master->progress("replica delay exceeds limit ($sync_delay);"
                  . " sleeping $sleep_time seconds [first pass]");
            sleep $sleep_time;
            ($sid, $diff)
              = $ldap_master->replica_seconds_behind_master($ldap_replica);
        }

        my ($result, $fmsg);
        $result = ($diff <= $sync_delay);
        my $comparison;
        if ($result) {
            $comparison = 'less';
        } else {
            $comparison = 'more';
        }
        $fmsg = "replica is $diff seconds behind master, "
          . "$comparison than limit of $sync_delay (sid: $sid)";

        $flextests->add($result, $fmsg);
        if (!$result) {
            $ldap_master->problem(
                "replica delay exceeds limit ($sync_delay) [final pass]");
        }
    }

    return $flextests;
}

=item datetimez_to_epoch

Given a string in "datetimez" format, return the corresponding epoch time.
Example: '20180608141044' yeilds 1528492244.

=cut

sub datetimez_to_epoch {
    my ($datetimez) = @_;

    if (!defined($datetimez)) {
        croak 'missing parameter for datetimez_to_epoch';
    }

    if ($datetimez =~ m{^(\d{4})(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)}ixsm) {
        my $year   = $1;
        my $month  = $2;
        my $dom    = $3;
        my $hour   = $4;
        my $minute = $5;
        my $sec    = $6;

        my $epoch = timelocal($sec, $minute, $hour, $dom, $month - 1, $year);
        return $epoch;
    } else {
        croak "cannot parse datetime '$datetimez'";
    }
}

=item get_contextCSNs

Returns a hash with keys the SID's (e.g., '064', '065') and values
the array ref [ datetime, epochtime].

Example:
  (
    '064' => ['20180608141044', 1528492244],
    '065' => ['20180608141044', 1528492244]
  )

=cut

sub get_contextCSNs {
    my ($self) = shift;

    my ($base) = @_;

    $self->progress(
        "about to get contextCSNs for $base on " . $self->server());

    my $search_result = $self->ldap_search(
        'base'   => $base,
        'scope'  => 'base',
        'filter' => 'objectclass=*',
        'attrs'  => ['contextCSN'],
    );

    my %dn_to_attrs = %{$search_result};

    my @keys = keys %dn_to_attrs;

    if (!exists($dn_to_attrs{$base})) {
        $self->problem("no contextCSN for base '$base'");
        return ();
    }

    my %values = %{ $dn_to_attrs{$base} };
    if (!exists($values{'contextcsn'})) {
        $self->problem("no contextCSN for base '$base'");
        return ();
    }

    my @csns = @{ $values{'contextcsn'} };
    if (scalar(@csns) == 0) {
        $self->problem("no contextCSN for base '$base'");
        return ();
    }

    my %results = ();
    foreach my $csn (@csns) {
        $self->progress("processing $csn");
        if ($csn =~ m{^(\d{14})[.](\d+)Z[#]\d+[#](\d\d\d)[#]\d+$}ixsm) {
            my $datetime = $1;
            my $sid      = $3;
            $self->progress("sid is $sid");
            $self->progress("datetime is $datetime");
            my $epoch = datetimez_to_epoch($datetime);
            $results{$sid} = [$datetime, $epoch];
        } else {
            $self->problem("failed to parse CSN '$csn'");
            return ();
        }
    }

    return %results;
}

=item $self->test_sugal

Given a uid change a couple of that entry's person record attributes to
trigger an suGAL change. Then change them back.

1. Does a record exist corresponding to the supplied uid? If not, fail
test and exit.

2. Record the CURRENT values of the attributes we are changing. We
do this in order to restore them to their correct values at the end of
the test.

3. Change the suVisibIdentity setting to "stanford". This visibility
setting controls read access to the attributes we will be changing: cn,
suOtherName, and suDisplayNameLF. If don't change it, and it is set to
"none", then changing the attributes will have no effect and this test
accomplishes nothing.

4. Change the attributes to random values.

5. Check the values of the corresponding attribute to see if they have
changed to the new random value. If not, fail the test. E.g.: change
"cn" to a random value and check that suGALcn has the new value.

6. Restore the original values of the attributes we changed. ZZZ

=cut

sub test_sugal {    ## no critic (ProhibitExcessComplexity)
    my ($ldap_master) = shift;    # $self is a MASTER

    my ($uid, $delay) = @_;
    my $msg;

    my ($result, $fmsg);

    my $flextests = Stanford::FlexTest::Collection->new();
    $flextests->ok_mode(!$ldap_master->in_nagios_mode);

    # Set the number of times we loop waiting for everything to settle
    # down and the delay between each loop.
    if (!$delay) {
        $delay = 0;
    }
    my $number_of_tries       = 4;
    my $seconds_between_tries = 2 + $delay;

    my $DN;
    my ($original_people_entry, $random_people_entry);

    # Note that 'cn' and 'suOtherName' are multi-valued while
    # 'suDisplayNameLF' is single-valued.
    my @attributes       = qw/     cn      suothername sudisplaynamelf/;
    my @sugal_attributes = qw/sugalcn sugalsuothername sugalsudisplaynamelf/;

    # @extra_attributes are any extra attributes we want to save but not
    # set via $set_attributes_fref.
    my @extra_attributes = qw/suvisibidentity/;

    ## ### ###    ## ### ###    ## ### ###    ## ### ###    ## ### ###
    # Get the values for @attributes from $uid's people record. Returns a
    # hash of arrays refs.
    my $get_values_fref = sub {
        my @all_attributes
          = (@attributes, @sugal_attributes, @extra_attributes);
        my $results_href
          = $ldap_master->get_attributes_uid($uid, \@all_attributes);

        # We only want the people part.
        my $dn1;
        foreach my $entry (keys %{$results_href}) {
            if ($entry =~ m{^suRegID=}xsm) {
                $dn1 = $entry;
                return ($dn1, $results_href->{$dn1});
            }
        }
    };
    ## ### ###    ## ### ###    ## ### ###    ## ### ###    ## ### ###
    my $to_sugal_fref = sub {
        my ($attribute) = @_;
        return 'sugal' . $attribute;
    };
    ## ### ###    ## ### ###    ## ### ###    ## ### ###    ## ### ###
    my $set_visibility = sub {
        my ($visib_value) = @_;
        $ldap_master->change_attribute($DN, 'suVisibIdentity', [$visib_value]);
    };
    ## ### ###    ## ### ###    ## ### ###    ## ### ###    ## ### ###
    # Given a hash reference that maps attributes to values, set the
    # attribute values for $DN to these attributes. If there is an
    # attribute missing from the hash reference, delete that attribute.
    my $set_attributes_fref = sub {
        my ($people_entry_to_change_to) = @_;

        foreach my $attribute (@attributes) {

            my $values_aref = $people_entry_to_change_to->{$attribute};

            if ($values_aref) {
                if (scalar(@{$values_aref}) > 0) {
                    my @values_current = @{$values_aref};
                    my $values_formatted = join(q{,}, @values_current);
                    $msg
                      = "changing attribute '$attribute' to '$values_formatted'";
                    $ldap_master->progress($msg);
                    $ldap_master->change_attribute($DN, $attribute,
                        $values_aref);
                } else {
                    $ldap_master->delete_attributes($DN, $attribute);
                    $ldap_master->progress("deleting attribute '$attribute'");
                }
            } else {
                $ldap_master->delete_attributes($DN, $attribute);
                $ldap_master->progress("deleting attribute '$attribute'");
            }
        }
        return;
    };
    ## ### ###    ## ### ###    ## ### ###    ## ### ###    ## ### ###
    ## ### ###    ## ### ###    ## ### ###    ## ### ###    ## ### ###
    # Step 1. Check that there is a record corresponding to uid. If not,
    # fail the test.
    ($DN, $original_people_entry) = $get_values_fref->();

    $result = $DN;
    $fmsg   = "found record corresponding to $uid";
    $flextests->add($result, $fmsg);

    # If the above test failed, there is no reason to continue.
    if (!$result) {
        return $flextests;
    }
    $ldap_master->progress("DN is '$DN'");
    $ldap_master->progress(Dumper $original_people_entry);

    # Step 2. Note that $original_people_entry has the original values of
    # the attributes we are saving, so Step 2 is already done.

    # Step 3. Change the value of suVisibIdentity to "stanford" so changes
    # will propagate to the suGAL attributes.
    my $visibility_original = $original_people_entry->{'suvisibidentity'}->[0];
    $msg = "original value of  suVisibIdentity is '$visibility_original'";
    $ldap_master->progress($msg);

    $set_visibility->('none');
    $ldap_master->progress(q{set suVisibIdentity to 'stanford'});

    # 4. Change the attributes to random values.

    # We want a copy of the original people data but with the first
    # attribute of each multi-valued attribute set to something random.
    $random_people_entry = Clone::clone($original_people_entry);

    foreach my $attribute (@attributes) {
        my $values_aref = $original_people_entry->{$attribute};

        my @values;
        if (!$values_aref) {
            $ldap_master->progress("no values for attribute '$attribute'");
            @values = ();
        } else {
            @values = @{ $original_people_entry->{$attribute} };
        }

        if (scalar(@values) > 0) {
            $ldap_master->progress("randomizing a '$attribute' value");
            $values[0] = Stanford::LDAP::RegressionTesting::gen_random(10);
            $random_people_entry->{$attribute} = \@values;
        } else {
            $ldap_master->progress("uid $uid has no '$attribute' values");
        }
    }

    $ldap_master->progress('-------------------------------------');
    $ldap_master->progress('Randomized entry is:');
    $ldap_master->progress(Dumper $random_people_entry);
    $ldap_master->progress('-------------------------------------');

    # Change the values in the Directory to $random_people_entry.
    $set_attributes_fref->($random_people_entry);

    # 5. Read the current attributes and make sure they match the original
    # values.

    ## ##     ## ##     ## ##     ## ##     ## ##     ## ##     ## ##
    # Given a DN and a hash reference to attribute values, verify that the
    # hash reference values match the ones corresponding to DN.
    my $test_sugal_matches_fref = sub {
        my ($dn, $people_entry) = $get_values_fref->();
        foreach my $attribute (@attributes) {
            my $values       = $people_entry->{$attribute};
            my $sugal_values = $people_entry->{ $to_sugal_fref->($attribute) };

            # Make sure the arrays are sorted.
            my @values_sorted       = ();
            my @sugal_values_sorted = ();
            if ($values) {
                @values_sorted = sort (@{$values});
            }
            if ($sugal_values) {
                @sugal_values_sorted = sort (@{$sugal_values});
            }

            $msg = "attribute '$attribute' is " . join(q{,}, @values_sorted);
            $ldap_master->progress($msg);
            $msg = "attribute 'suGal$attribute' is "
              . join(q{,}, @sugal_values_sorted);
            $ldap_master->progress($msg);
            my $result1
              = Stanford::LDAP::RegressionTesting::compare_arrays(
                \@values_sorted, \@sugal_values_sorted);
            if (!$result1) {
                return 0;
            }
        }
        return 1;
    };
    ## ##     ## ##     ## ##     ## ##     ## ##     ## ##     ## ##

    #<<<  do not let perltidy touch this
    $result = $ldap_master->try_several_times(
        $test_sugal_matches_fref,
        $number_of_tries,
        $seconds_between_tries
    );
    #>>>
    $fmsg = 'suGAL syncing matched';
    $flextests->add($result, $fmsg);

    # Step 6. Change the values in the Directory back to the original
    # values. We have to do a separate restoration of suVisibIdentity.
    $set_attributes_fref->($original_people_entry);
    $ldap_master->progress('restored attributes to original values');

    $set_visibility->($visibility_original);
    $msg = "set suVisibIdentity back to original value '$visibility_original'";
    $ldap_master->progress($msg);

    # Check that the suGAL attributes match.
    #<<<  do not let perltidy touch this
    $result = $ldap_master->try_several_times(
        $test_sugal_matches_fref,
        $number_of_tries,
        $seconds_between_tries
    );
    #>>>
    $fmsg = 'suGAL syncing matched after original values reset';
    $flextests->add($result, $fmsg);

    # Check that the original values are correct.
    ## ##     ## ##     ## ##     ## ##     ## ##     ## ##     ## ##
    my $test_original_matches_fref = sub {
        my ($dn, $people_entry) = $get_values_fref->();
        foreach my $attribute (@attributes) {
            my $values          = $people_entry->{$attribute};
            my $original_values = $original_people_entry->{$attribute};
            my $result1
              = Stanford::LDAP::RegressionTesting::compare_arrays($values,
                $original_values);
            if (!$result1) {
                return 0;
            }
        }
        return 1;
    };
    ## ##     ## ##     ## ##     ## ##     ## ##     ## ##     ## ##
    # Check that the suGAL attributes match.
    my $orig_rv = (
        $ldap_master->try_several_times(
            $test_original_matches_fref, $number_of_tries,
            $seconds_between_tries
        )
    );

    $result = $orig_rv;
    $fmsg   = 'original values were reset correctly';
    $flextests->add($result, $fmsg);

    if (!$orig_rv) {
        print 'WARNING: your original values were not reset correctly!'
          . "You will have to reset them yourself manually. Your original attributes:\n";
        foreach my $attribute (keys %{$original_people_entry}) {
            print "$attribute: " . $original_people_entry->{$attribute} . "\n";
        }
    }

    return $flextests;
}

=item $self->test_sync_privgroup

This tests the service that syncs suPrivilegeGroup's from the people tree
to the accounts tree.

It adds a suPrivilegeGroup attribute to a test account's people tree,
verifies that the attribute shows up in the accounts tree, deletes it from
the people tree and verifies that it gets deleted from the accounts tree.

This method needs to be passed a uid and suPrivilegeGroup. The uid
indicates which user to run the test on and the suPrivilegeGroup will be
the one added and removed.

=cut

sub test_sync_privgroup {
    my ($ldap) = shift;    # $self is a MASTER

    my ($uid, $supg) = @_;

    my ($result, $fmsg);

    my $flextests = Stanford::FlexTest::Collection->new();
    $flextests->ok_mode(!$ldap->in_nagios_mode);

    ## Step 1. Get the people-tree DN corresponding to this uid.
    my $dn = $ldap->get_dn_from_uid($uid, 'people');

    $result = $dn;
    $fmsg   = "found record corresponding to $uid";
    $flextests->add($result, $fmsg);

    # If the above test failed, there is no reason to continue.
    if (!$result) {
        return $flextests;
    }
    $ldap->progress("DN is '$dn'");

    # Store the current state of whether this user is in
    # $supg. We will need this at the end.
    my $in_privgroup = $ldap->in_suPrivilegeGroup($uid, $supg, 'people');

    ##############################################################
    # Local functions to test privgroup membership               #
    ##############################################################
    # ## ##    # ## ##    # ## ##    # ## ##    # ## ##    # ## ##
    my $in_privgroup_people_fref = sub {
        return $ldap->in_suPrivilegeGroup($uid, $supg, 'people');
    };

    # ## ##    # ## ##    # ## ##    # ## ##    # ## ##    # ## ##
    my $not_in_privgroup_people_fref = sub {
        return !$ldap->in_suPrivilegeGroup($uid, $supg, 'people');

    };

    # ## ##    # ## ##    # ## ##    # ## ##    # ## ##    # ## ##
    my $in_privgroup_accounts_fref = sub {
        return $ldap->in_suPrivilegeGroup($uid, $supg, 'accounts');

    };

    # ## ##    # ## ##    # ## ##    # ## ##    # ## ##    # ## ##
    my $not_in_privgroup_accounts_fref = sub {
        return !$ldap->in_suPrivilegeGroup($uid, $supg, 'accounts');

    };

    # ## ##    # ## ##    # ## ##    # ## ##    # ## ##    # ## ##

    # Test 1. Remove $supg from the PEOPLE tree and verify.
    $ldap->remove_user_from_privgroup($dn, $supg);

    $result = $ldap->try_several_times($not_in_privgroup_people_fref, 4, 2);
    $fmsg = "$uid not in $supg/people (initial)";
    $flextests->add($result, $fmsg);

    # Test 2. Verify that the ACCOUNTS tree does not think user is in $supg.
    $result = $ldap->try_several_times($not_in_privgroup_accounts_fref, 4, 2);
    $fmsg = "$uid not in $supg/accounts (initial)";
    $flextests->add($result, $fmsg);

    # Test 3. Add user to $supg on the PEOPLE tree and verify.
    $ldap->add_user_to_privgroup($dn, $supg);

    $result = $ldap->try_several_times($in_privgroup_people_fref, 4, 2);
    $fmsg = "$uid is now in $supg/people";
    $flextests->add($result, $fmsg);

    # Test 4. Verify that the ACCOUNTS tree now thinks user is in $supg.
    $result = $ldap->try_several_times($in_privgroup_accounts_fref, 4, 2);
    $fmsg = "$uid is now in $supg/accounts";
    $flextests->add($result, $fmsg);

    # Test 5. Remove $supg from the people tree and verify that the
    # PEOPLE tree does not think $uid is $supg.
    $ldap->remove_user_from_privgroup($dn, $supg);

    $result = $ldap->try_several_times($not_in_privgroup_people_fref, 4, 2);
    $fmsg = "$uid not in $supg/people (second removal)";
    $flextests->add($result, $fmsg);

    # Test 6. Verify that the sync service removed $uid from $supg
    # relative to the ACCOUNTS tree.
    $result = $ldap->try_several_times($not_in_privgroup_accounts_fref, 4, 2);
    $fmsg = "$uid not in $supg/accounts (second removal)";
    $flextests->add($result, $fmsg);

    # Restore membership to what it was before we started this
    # function.
    if ($in_privgroup) {

        # Add
        $ldap->add_user_to_privgroup($dn, $supg);

        # Test 7A. Verify that $uid is in $supg relative to PEOPLE.
        $result = $ldap->try_several_times($in_privgroup_people_fref, 4, 2);
        $fmsg = "${uid}'s $supg status restored to original state (people)";
        $flextests->add($result, $fmsg);

        # Test 8A. Verify that $uid is in $supg relative to ACCOUNTS.
        $result = $ldap->try_several_times($in_privgroup_accounts_fref, 4, 2);
        $fmsg = "${uid}'s $supg status restored to original state (accounts)";
        $flextests->add($result, $fmsg);
    } else {

        # Remove
        $ldap->remove_user_from_privgroup($dn, $supg);

        # Test 7B. Verify that $uid is _not_ in $supg relative to PEOPLE.
        $result
          = $ldap->try_several_times($not_in_privgroup_people_fref, 4, 2);
        $fmsg = "${uid}'s $supg status restored to original state (people)";
        $flextests->add($result, $fmsg);

        # Test 8b. Verify that $uid is _not_ in $supg relative to ACCOUNTS.
        $result
          = $ldap->try_several_times($not_in_privgroup_accounts_fref, 4, 2);
        $fmsg = "${uid}'s $supg status restored to original state (accounts)";
        $flextests->add($result, $fmsg);
    }

    return $flextests;
}

=back

=cut

##########################################################

__PACKAGE__->meta->make_immutable;

1;

__END__
